﻿// Создал Андрей Котов.

using System;
using System.Linq;
using System.Linq.Expressions;
using System.Collections.Generic;
using System.Reflection;
using System.Net.Mail;
using System.IO;
using System.Xml;
using System.Globalization;
using CrossCompileSharing.Data;
using CrossCompileSharing.Log;
using CrossCompileSharing.UiStrings;

namespace BillAnalyserWebApp.Code.Data
{
	[Obfuscation(Exclude = true)]
	public class SmtpParameters
	{
		public string Host { get; set; }
		public int Port { get; set; }
		public string User { get; set; }
		public string Subject { get; set; }

		// it3xl.ru: Передача пароля на клиет запрещена!
		//public string Pass { get; set; }
	}

	[Obfuscation(Exclude = true)]
	public class PostResult
	{
		public string Abonent { get; set; }
		public string Email { get; set; }
		/// <summary>
		/// Признак, что перед отправкой этого сообщения eMail была прервана рассылка почты.
		/// </summary>
		public bool Terminated { get; set; }
		/// <summary>
		/// Используется для отображения на клиенте, как прошла операция отправки почты. Рашньше называлось ErrorMessage, но пришлось переименовать,
		///  т.к. тут передаешься еще сообщение об успешной операции.
		/// </summary>
		public string OperationResultMessage { get; set; }
	}

	internal class OperationState
	{
		public int CountPosted { get; set; }

		public int NumberPosts { get; set; }

		public Boolean Terminated { get; set; }

		private DateTime _created = DateTime.Now;
		public DateTime Created
		{
			get { return _created; }
		}

		public OperationState GetCopy()
		{
			return new OperationState
			{
				CountPosted = CountPosted,
				NumberPosts = NumberPosts,
				Terminated = Terminated,
				_created = _created,
			};
		}
	}

	/// <summary>
	/// Класс взаимодействия с почтовой службой.
	/// </summary>
	public static class Smtp
	{
		private static readonly Dictionary<String, OperationState> _operationsSate = new Dictionary<String, OperationState>();

		internal static Boolean SetOperationStateTry(String stateKey, OperationState operationState)
		{
			lock (_operationsSate)
			{
				if (_operationsSate.ContainsKey(stateKey))
				{

					return false;
				}

				_operationsSate.Add(stateKey, operationState);
			}

			return true;
		}

		/// <summary>
		/// Возращает точную копию для чтения, чтоб не создавать зависимости в многопоточности.
		/// </summary>
		/// <param name="stateKey"></param>
		/// <returns></returns>
		internal static OperationState GetOperationStateCloneTry(String stateKey)
		{
			lock (_operationsSate)
			{
				if (_operationsSate.ContainsKey(stateKey) == false)
				{
					return null;
				}

				return _operationsSate[stateKey].GetCopy();
			}
		}

		internal static void ReleaseOperationStateTry(String stateKey)
		{
			lock (_operationsSate)
			{
				if (_operationsSate.ContainsKey(stateKey))
				{
					_operationsSate.Remove(stateKey);
				}

				// it3xl.ru: Защитимся от любых утечек памяти каждый раз удаляя вообще все устаревшие состояния.
				// Если за четыре дня у вас не отправились все письма в рассылке, то ... вам не повезло, т.к. в малых и средних предприятиях численность до 100(150) человек.
				const int MaxMailingPeriodDays = 4;
				var terminatedStateKeys = from terminated in _operationsSate where terminated.Value.Created < DateTime.Now.AddDays(-MaxMailingPeriodDays) select terminated.Key;
				foreach (var key in terminatedStateKeys)
				{
					_operationsSate.Remove(key);
				}

			}
		}

		public static void IncreaseOperationStateProcessedCount(string stateKey)
		{
			lock (_operationsSate)
			{
				if (_operationsSate.ContainsKey(stateKey))
				{
					_operationsSate[stateKey].CountPosted++;
				}
			}
		}

		public static void TerminatePostBills(String stateKey)
		{
			lock (_operationsSate)
			{
				if (_operationsSate.ContainsKey(stateKey))
				{
					_operationsSate[stateKey].Terminated = true;
				}
			}
		}

		/// <summary>
		/// Отправка сообщения
		/// </summary>
		/// <param name="abonent"></param>
		/// <param name="period"></param>
		/// <param name="smtpClient"></param>
		/// <param name="useCostsForServices">Отображать в письме "Затраты по услугам".</param>
		/// <param name="useDetalization">Отображать в письме "Детализацию".</param>
		/// <returns></returns>
		public static ServiceResult<PostResult> SendMessage(Abonent abonent, DateTime period, Boolean useCostsForServices, Boolean useSubtotal, Boolean useDetalization, Calls.ViewModFor viewModFor, SmtpParameters smtpParameters, SmtpClient smtpClient, String stateKey)
		{
			if (abonent == null)
			{
				return new ServiceResult<PostResult>(UiMessages.Данные_по_абоненту_не_проинициализированы);
			}

			var result = new ServiceResult<PostResult>
			{
				DataOrNullOrDefValue = new PostResult
				{
					Abonent = abonent.Name,
					Email = abonent.Email,
				},
			};

			if (CommonFunctions.IsEmail(abonent.Email) == false)
			{
				result.BusinessErrorNotNulled.Add(UiMessages.Почтовый_адрес_абонента_не_корректен);
				result.DataOrNullOrDefValue.OperationResultMessage = result.BusinessErrorNotNulled.GetAllErrorsText();

				return result;
			}

			var dataPost = GetDataForPost(abonent, period, viewModFor);
			if (dataPost.HasErrors)
			{
				result.BusinessErrorNotNulled.Add(UiMessages.Данные_для_бланка_неверные);
				result.DataOrNullOrDefValue.OperationResultMessage = result.BusinessErrorNotNulled.GetAllErrorsText();

				return result;
			}

			var bodyResult = GenerateBodyMessage(dataPost.DataOrNullOrDefValue, useCostsForServices, useSubtotal, useDetalization);

			if (bodyResult.HasErrors
				|| string.IsNullOrEmpty(bodyResult.DataOrNullOrDefValue))
			{
				result.BusinessErrorNotNulled.Add(UiMessages.Не_удалось_создать_текст_сообщения);
				result.DataOrNullOrDefValue.OperationResultMessage = result.BusinessErrorNotNulled.GetAllErrorsText();

				return result;
			}

			MailMessage message;
			// it3xl.ru: Задание адреса From в сообщении обязательно, иначе не работает отправка
			var fromAdressStr = "(unknown)";
			try
			{
				MailAddress fromAdress = null;
				// it3xl.ru: Имя пользователя для доступа к SMTP-серверу может быть записано не в формате e-mail, что вызовет ошибку.
				try
				{
					fromAdress = new MailAddress(smtpParameters.User);
				}
				catch { }

				// it3xl.ru: Если адрес From не удалось задать, то зададим его принудительно, т.к. если не будет задан параметр GUI Пользователь в формате eMail,
				//  то на многих почтовых серверах отправка не пройдет.
				if (fromAdress == null)
				{
					fromAdress = new MailAddress("BAnalyzer@mts.ru");
				}

				message = new MailMessage { From = fromAdress };
				fromAdressStr = fromAdress.Address;

				message.To.Add(new MailAddress(abonent.Email));
				message.Subject = smtpParameters.Subject;
				message.BodyEncoding = System.Text.Encoding.UTF8;
				message.IsBodyHtml = true;
			}
			catch (Exception ex)
			{
				// it3xl.ru: Включил в информирование текст исключения, чтоб клиент сам мог разобраться, что у него не работает без обращения в поддержку.
				var createMessageError = String.Format(UiMessages.Ошибка_создания_почтового_сообщения__, ex.Message);
				Logger.Write(ex, createMessageError);
				result.BusinessErrorNotNulled.Add(createMessageError);
				result.DataOrNullOrDefValue.OperationResultMessage = result.BusinessErrorNotNulled.GetAllErrorsText();

				return result;
			}

			message.Body = bodyResult.DataOrNullOrDefValue;

			if (GetOperationStateCloneTry(stateKey).Terminated == false)
			{
				try
				{
					smtpClient.Send(message);
				}
				catch (Exception ex)
				{
					// it3xl.ru: Включил в информирование текст исключения, чтоб клиент сам мог разобраться, что у него не работает без обращения в поддержку.
					var sendErrorMessage = String.Format(UiMessages.Ошибка_отправки_почтового_сообщения__From, ex.Message, fromAdressStr);
					Logger.Write(ex, sendErrorMessage);
					result.BusinessErrorNotNulled.Add(sendErrorMessage);
					result.DataOrNullOrDefValue.OperationResultMessage = result.BusinessErrorNotNulled.GetAllErrorsText();

					return result;
				}
				result.DataOrNullOrDefValue.OperationResultMessage = UiMessages.Отправлено_успешно;
			}
			else
			{
				result.DataOrNullOrDefValue.Terminated = true;
				result.DataOrNullOrDefValue.OperationResultMessage = UiMessages.Отправка_почты_прервана_пользователем_;
			}

			return result;
		}

		/// <summary>
		/// it3xl.ru: В XmlDocument для XML не работает GetElementById. Это его аналог для атрибута id в XML.
		/// Регистр важен.
		/// </summary>
		/// <param name="doc"></param>
		/// <param name="id"></param>
		/// <returns></returns>
		public static XmlElement GetElementByIdInXml(this XmlDocument doc, String id)
		{
			const String QueryById = "//*[@id='{0}']";
			return (XmlElement)doc.SelectSingleNode(String.Format(QueryById, id));
		}

		/// <summary>
		/// it3xl.ru: Получение элемента по значению атрибута "class".
		/// Регистр важен.
		/// </summary>
		/// <param name="doc"></param>
		/// <param name="id"></param>
		/// <returns></returns>
		public static XmlElement GetElementByClassInXml(this XmlDocument doc, String id)
		{
			const String QueryById = "//*[@class='{0}']";
			return (XmlElement)doc.SelectSingleNode(String.Format(QueryById, id));
		}

		/// <summary>
		/// Генерация тела сообщения
		/// </summary>
		/// <param name="dataPost"></param>
		/// <returns></returns>
		private static ServiceResult<String> GenerateBodyMessage(DataForPost dataPost, Boolean useCostsForServices, Boolean useSubtotal, Boolean useDetalization)
		{
			if (SmtpMailTemplate.TemplateExist == false)
			{
				return new ServiceResult<string>(String.Format(UiMessages.Шаблон_почтового_сообщения_отсутствует__, SmtpMailTemplate.TemplateFilePath));
			}

			var xmlDocument = new XmlDocument();
			try
			{
				xmlDocument.LoadXml(SmtpMailTemplate.Value);
			}
			catch (Exception ex)
			{
				Logger.Write(ex, "Ошибка формирования тела сообщения из шаблона XML.");
				return new ServiceResult<string>(ServiceResult.AddBusinessErrorWithDefaultError.Yes);
			}

			try
			{
				// заменям шабон на дату
				XmlNode tdPeriod = xmlDocument.GetElementByIdInXml("period");
				if (tdPeriod != null)
				{
					tdPeriod.InnerXml = tdPeriod.InnerXml.Replace(
						"[mm-yy]",
						String.Format(
							"{0} {1}",
							(new CultureInfo("Ru-RU", true)).DateTimeFormat.GetMonthName(dataPost.period.Month),
							dataPost.period.Year.ToString()
						)
					);
				}
				// Вставляем имя абонента
				XmlNode tdAbonent = xmlDocument.GetElementByIdInXml("abonent");
				if(tdAbonent!=null)
				{
					tdAbonent.InnerXml = tdAbonent.InnerXml.Replace("[abonent]", dataPost.Abonent);
				}

				// Обрабатываем телефоны
				XmlNode templatePhoneElement = xmlDocument.GetElementByClassInXml("phone_root");
				if (templatePhoneElement != null)
				{
					var templateElementParent = templatePhoneElement.ParentNode;
					templateElementParent.RemoveChild(templatePhoneElement);

					foreach (var t in dataPost.lists)
					{
						var currentPhoneElement = templatePhoneElement.Clone();
						FillTable(currentPhoneElement, t, useCostsForServices, useSubtotal, useDetalization);
						templateElementParent.AppendChild(currentPhoneElement);
					}
				}

				// Редактируем таблицу "Итого по абоненту" 
				XmlNode tableResultAbonent = xmlDocument.GetElementByIdInXml("table_result_all");
				if (tableResultAbonent != null)
				{
					var chargeAbonent = String.Format("{0:f2}", dataPost.Charge);
					tableResultAbonent.InnerXml = tableResultAbonent.InnerXml.Replace("[result_abonent_charge]", chargeAbonent);
					var compensationAbonent = String.Format("{0:f2}", dataPost.Compensation);
					tableResultAbonent.InnerXml = tableResultAbonent.InnerXml.Replace("[result_abonent_compensation]", compensationAbonent);
					var retantionAbonent = String.Format("{0:f2}", dataPost.Retantion);
					tableResultAbonent.InnerXml = tableResultAbonent.InnerXml.Replace("[result_abonent_retantion]", retantionAbonent);
				}

				using (var sw = new StringWriter())
				{
					var tx = new XmlTextWriter(sw);
					xmlDocument.WriteTo(tx);

					return new ServiceResult<string> { DataOrNullOrDefValue = sw.ToString() };
				}
			}
			catch (Exception ex)
			{
				Logger.Write(ex);

				return new ServiceResult<string>(ServiceResult.AddBusinessErrorWithDefaultError.Yes);
			}
		}

		/// <summary>
		/// Получение данных для формирования бланка
		/// </summary>
		/// <param name="abonent"></param>
		/// <param name="period"></param>
		/// <returns></returns>
		private static ServiceResult<DataForPost> GetDataForPost(Abonent abonent, DateTime period, Calls.ViewModFor viewModFor)
		{
			var dataPost = new DataForPost
			{
				AbonentId = abonent.Id,
				Abonent = abonent.Name,
				period = period,
				lists = new List<DataForPostByPhone>(),
			};

			var userPhonesResult = Phones.UserPhones(abonent.Id);
			if (userPhonesResult.HasErrors)
			{
				// !!!
				return new ServiceResult<DataForPost>(userPhonesResult.BusinessErrorNotNulled);
			}

			foreach (var phone in userPhonesResult.DataOrNullOrDefValue)
			{
				var listCallsResult = Calls.CallGetByPhoneId(phone.Id, period, viewModFor);
				if (listCallsResult.HasErrors)
				{
					// !!!
					return new ServiceResult<DataForPost>(listCallsResult.BusinessErrorNotNulled);
				}

				var listServicesResult = BillServices.BillServicesGetByPeriod(phone.Id, period);
				if (listServicesResult.HasErrors)
				{
					// !!!
					return new ServiceResult<DataForPost>(listServicesResult.BusinessErrorNotNulled);
				}

				dataPost.lists.Add(new DataForPostByPhone
				{
					phoneId = phone.Id,
					Phone = phone.PhoneStr,
					listCalls = listCallsResult.DataOrNullOrDefValue,
					listServices = listServicesResult.DataOrNullOrDefValue,
				});
			}

			return new ServiceResult<DataForPost>
			{
				DataOrNullOrDefValue = dataPost,
			};
		}

		/// <summary>
		/// Заполнение таблиц по телефону
		/// </summary>
		/// <param name="tableByPhoneSource"></param>
		/// <param name="dataPhone"></param>
		private static void FillTable(XmlNode tableByPhoneSource, DataForPostByPhone dataPhone, Boolean useCostsForServices, Boolean useSubtotal, Boolean useDetalization)
		{
			// Заменяем шаблон номера телефона на номер телефона
			tableByPhoneSource.InnerXml = tableByPhoneSource.InnerXml.Replace("[abonent_phone]", dataPhone.Phone);

			var xpath = @"//*[@id='table_calls']";
			var tableCalls = tableByPhoneSource.SelectSingleNode(xpath);
			if (tableCalls != null)
			{
				if (useDetalization)
				{
					FillTableCalls(tableCalls, dataPhone.listCalls);
				}
				else
				{
					tableCalls.ParentNode.RemoveChild(tableCalls);
				}
			}

			xpath = @"//*[@id='table_services']";
			var tableServices = tableByPhoneSource.SelectSingleNode(xpath);
			if (tableServices != null)
			{
				if (useCostsForServices)
				{
					FillTableServices(tableServices, dataPhone.listServices);
				}
				else
				{
					tableServices.ParentNode.RemoveChild(tableServices);
				}
			}

			xpath = @"//*[@id='table_result']";
			var tableResult = tableByPhoneSource.SelectSingleNode(xpath);
			if (tableResult != null)
			{
				if (useSubtotal)
				{
					FillTableResult(tableResult, dataPhone);
				}
				else
				{
					tableResult.ParentNode.RemoveChild(tableResult);
				}
			}
		}

		/// <summary>
		/// Если текст пустой, то поставить пробел.
		/// Нужно, чтоб не схлопывались ячейки таблиц при отображении HTML.
		/// </summary>
		/// <param name="xmlElement"></param>
		/// <param name="text"></param>
		public static void SetInnerTextOrGap(this XmlElement xmlElement, String text)
		{
			if (String.IsNullOrEmpty(text))
			{
				xmlElement.InnerXml = "&nbsp;";
			}
			else
			{
				xmlElement.InnerText = text;
			}
		}

		/// <summary>
		/// Заполнение таблицы звонков
		/// </summary>
		/// <param name="tableCalls"></param>
		/// <param name="listCalls"></param>
		private static void FillTableCalls(XmlNode tableCalls, IEnumerable<Call> listCalls)
		{
			foreach (var call in listCalls)
			{
				var newRow = tableCalls.OwnerDocument.CreateElement(@"tr");

				var newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.CallDate.ToString("dd.MM.yyyy HH:mm"));
				newRow.AppendChild(newCol);

				newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.PhoneB);
				newRow.AppendChild(newCol);

				newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.ZonePS);
				newRow.AppendChild(newCol);

				newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.ZoneVTK);
				newRow.AppendChild(newCol);

				newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.ServiceName);
				newRow.AppendChild(newCol);

				newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.Dur);
				newRow.AppendChild(newCol);

				newCol = tableCalls.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(call.CostNotVat.ToString());
				newRow.AppendChild(newCol);

				tableCalls.AppendChild(newRow);
			}
		}

		/// <summary>
		/// Заполнение таблицы услуг
		/// </summary>
		/// <param name="tableServices"></param>
		/// <param name="listServices"></param>
		private static void FillTableServices(XmlNode tableServices, IEnumerable<BillService> listServices)
		{
			foreach (var service in listServices)
			{
				var newRow = tableServices.OwnerDocument.CreateElement(@"tr");

				var newCol = tableServices.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(service.Name);
				newRow.AppendChild(newCol);

				newCol = tableServices.OwnerDocument.CreateElement("td");
				newCol.InnerText = service.Duration.HasValue ? service.Duration.Value.ToString() : "0.0";
				newRow.AppendChild(newCol);

				newCol = tableServices.OwnerDocument.CreateElement("td");
				newCol.SetInnerTextOrGap(service.DurationUnit);
				newRow.AppendChild(newCol);

				newCol = tableServices.OwnerDocument.CreateElement("td");
				newCol.InnerText = service.SumAdd.HasValue ? service.SumAdd.Value.ToString() : "0.0";
				newRow.AppendChild(newCol);

				tableServices.AppendChild(newRow);
			}
		}


		/// <summary>
		/// Заполнение таблицы "итого"
		/// </summary>
		/// <param name="tableResult"></param>
		/// <param name="dataPhone"></param>
		private static void FillTableResult(XmlNode tableResult, DataForPostByPhone dataPhone)
		{
			string charges = string.Format("{0:f2}", dataPhone.Charges);
			tableResult.InnerXml = tableResult.InnerXml.Replace("[number_charge]", charges);
			string compensation = string.Format("{0:f2}", dataPhone.Compensation);
			tableResult.InnerXml = tableResult.InnerXml.Replace("[number_compensation]", compensation);
			string retantion = string.Format("{0:f2}", dataPhone.Retantion);
			tableResult.InnerXml = tableResult.InnerXml.Replace("[number_retantion]", retantion);
		}

		/// <summary>
		/// Вернет процент выполнения или 100%, если указанное состояние не найдено.
		/// </summary>
		/// <param name="stateKey"></param>
		/// <returns></returns>
		public static int GetOperationPercent(string stateKey)
		{
			var operationState = GetOperationStateCloneTry(stateKey);
			if (operationState == null)
			{
				// Если состояния нет, то считается, что оно завершено.
				return 100;
			}

			if (operationState.NumberPosts == 0)
			{
				return 0;
			}

			return operationState.CountPosted * 100 / operationState.NumberPosts;
		}
	}
}