﻿using System;
using System.Text.RegularExpressions;
using CrossCompileSharing.Data;
using CrossCompileSharing.UiStrings;
using Oracle.DataAccess.Client;

namespace BillAnalyserWebApp.Code.IFlex.DB.Oracle
{
	public static class BusinessErrorManager
	{
		public static BusinessError ExtractIfExist(Exception ex)
		{
			var businesErrorMessage = GetBusinesErrorMessage(ex);
			return new BusinessError(businesErrorMessage);
		}

		/// <summary>
		/// Скрытие от клиента ошибок базы данных, которые он не дожен видеть. Отключено, т.к. проект корпоративный и просмотор логов не сделан.
		/// </summary>
		/// <param name="ex"></param>
		/// <returns></returns>
		public static BusinessError ExtractIfExistOrMakeAsDefaultDbError(Exception ex)
		{
			var businesErrorMessage = GetBusinesErrorMessage(ex);
			if(String.IsNullOrEmpty(businesErrorMessage))
			{
				if (GetIsOracleDbConnectionFailure(ex.Message))
				{
					businesErrorMessage = UiMessages.Отсутствует_связь_с_базой_данных_Oracle;
				}
				else if (GetNeedRestartByOracle(ex.Message))
				{
					businesErrorMessage = UiMessages.Отсутствует_связь_с_базой_данных_Oracle;
//                    try
//                    {
//                        OracleConnection.ClearAllPools();
//                    }
//// ReSharper disable EmptyGeneralCatchClause
//                    catch{}
//// ReSharper restore EmptyGeneralCatchClause
				}
				else
				{
					// it3xl.ru: Отключаем скрытие ошибки от пользователя.
					var turnoutErrorHiding = true;
					if(turnoutErrorHiding)
					{
						businesErrorMessage = ex.ToString();
					}
					else
					{
						businesErrorMessage = UiMessages.UnknownDbBusinessErrorMessage;
					}
				}
			}
			return new BusinessError(businesErrorMessage);
		}

		private static bool GetIsOracleDbConnectionFailure(string errorMessage)
		{
			return
				// ORA-12547: TNS:lost contact
				errorMessage.StartsWith("ORA-12547")
				// ORA-03135: connection lost contact
				|| errorMessage.StartsWith("ORA-03135")
				// ORA-12545: Connect failed because target host or object does not exist
				|| errorMessage.StartsWith("ORA-12545")
				;
		}

		/// <summary>
		/// Для восстановления работы нужно сбросить пул соединений с базой на стороне .NET Framework или перестартовать IIS.
		/// </summary>
		/// <param name="errorMessage"></param>
		/// <returns></returns>
		private static bool GetNeedRestartByOracle(string errorMessage)
		{
			return
				// ORA-12571: TNS:packet writer failure
				errorMessage.StartsWith("ORA-12571")
				// ORA-12518: TNS:listener could not hand off client connection
				|| errorMessage.StartsWith("ORA-12518")
				// ORA-12514: TNS:listener does not currently know of service requested in connect descriptor
				|| errorMessage.StartsWith("ORA-12514")
				;
		}

		private static string GetBusinesErrorMessage(Exception ex)
		{
			String businesErrorMessage = null;
			var match = Regex.Match(
				ex.Message,
				// ITretykaov: Так уж сложилось, что в Оракле так зашито (busines, а не business), видимо еще из Сибоса так пошло.
				"<busines_error>(.+?)</busines_error>",
				RegexOptions.IgnoreCase
				);
			if (0 < match.Captures.Count)
			{
				businesErrorMessage = match.Groups[1].Value;
			}
			return businesErrorMessage;
		}
	}
}